'use strict';
Object.defineProperty(exports, '__esModule', { value: true });
exports.qgamma = void 0;
var debug_1 = require('@mangos/debug');
var logger_1 = require('../../../packages/common/logger.cjs');
var r_func_1 = require('../../r-func.cjs');
var expm1_1 = require('../exp/expm1.cjs');
var dgamma_1 = require('./dgamma.cjs');
var lgammafn_sign_1 = require('../../special/gamma/lgammafn_sign.cjs');
var pgamma_1 = require('./pgamma.cjs');
var qnorm_1 = require('../normal/qnorm.cjs');
const printer_qchisq_appr = (0, debug_1.debug)('qchisq_appr');
function qchisq_appr(p, nu, g, lower_tail, log_p, tol) {
    const C7 = 4.67;
    const C8 = 6.66;
    const C9 = 6.73;
    const C10 = 13.32;
    let a;
    let ch;
    let p1;
    let p2;
    let q;
    let t;
    let x;
    if (isNaN(p) || isNaN(nu))
        return NaN;
    const rc = (0, logger_1.R_Q_P01_check)(log_p, p);
    if (rc !== undefined) {
        return rc;
    }
    if (nu <= 0) {
        return (0, logger_1.ML_ERR_return_NAN2)(printer_qchisq_appr, logger_1.lineInfo4);
    }
    const alpha = 0.5 * nu;
    const c = alpha - 1;
    if (nu < -1.24 * (p1 = (0, expm1_1.R_DT_log)(lower_tail, log_p, p))) {
        const lgam1pa = alpha < 0.5 ? (0, pgamma_1.lgamma1p)(alpha) : Math.log(alpha) + g;
        ch = Math.exp((lgam1pa + p1) / alpha + r_func_1.M_LN2);
        printer_qchisq_appr(' small chi-sq., ch0 = %d', ch);
    } else if (nu > 0.32) {
        x = (0, qnorm_1.qnorm)(p, 0, 1, lower_tail, log_p);
        p1 = 2 / (9 * nu);
        ch = nu * Math.pow(x * Math.sqrt(p1) + 1 - p1, 3);
        printer_qchisq_appr(' nu > .32: Wilson-Hilferty; x = %d', x);
        if (ch > 2.2 * nu + 6)
            ch = -2 * ((0, expm1_1.R_DT_Clog)(lower_tail, log_p, p) - c * Math.log(0.5 * ch) + g);
    } else {
        ch = 0.4;
        a = (0, expm1_1.R_DT_Clog)(lower_tail, log_p, p) + g + c * r_func_1.M_LN2;
        printer_qchisq_appr(' nu <= .32: a = %d', a);
        do {
            q = ch;
            p1 = 1 / (1 + ch * (C7 + ch));
            p2 = ch * (C9 + ch * (C8 + ch));
            t = -0.5 + (C7 + 2 * ch) * p1 - (C9 + ch * (C10 + 3 * ch)) / p2;
            ch -= (1 - Math.exp(a + 0.5 * ch) * p2 * p1) / t;
        } while (Math.abs(q - ch) > tol * Math.abs(ch));
    }
    return ch;
}
const printer_qgamma = (0, debug_1.debug)('qgamma');
const EPS1 = 0.01;
const EPS2 = 5e-7;
const EPS_N = 1e-15;
const MAXIT = 1000;
const pMIN = 1e-100;
const pMAX = 1 - 1e-14;
const i420 = 1 / 420;
const i2520 = 1 / 2520;
const i5040 = 1 / 5040;
function qgamma(p, alpha, scale, lower_tail, log_p) {
    let p_;
    let a;
    let b;
    let g;
    let ch;
    let p1;
    let p2;
    let s1;
    let s2;
    let s3;
    let s4;
    let s5;
    let t;
    let x;
    let i;
    let max_it_Newton = 1;
    let q = 0;
    if (isNaN(p) || isNaN(alpha) || isNaN(scale))
        return NaN;
    const rc = (0, logger_1.R_Q_P01_boundaries)(lower_tail, log_p, p, 0, Infinity);
    if (rc !== undefined) {
        return rc;
    }
    if (alpha < 0 || scale <= 0)
        return (0, logger_1.ML_ERR_return_NAN2)(printer_qgamma, logger_1.lineInfo4);
    if (alpha === 0)
        return 0;
    const fn_end = () => {
        x = 0.5 * scale * ch;
        if (max_it_Newton) {
            if (!log_p) {
                p = Math.log(p);
                log_p = true;
            }
            if (x === 0) {
                const _1_p = 1 + 1e-7;
                const _1_m = 1 - 1e-7;
                x = r_func_1.DBL_MIN;
                p_ = (0, pgamma_1.pgamma)(x, alpha, scale, lower_tail, log_p);
                if (lower_tail && p_ > p * _1_p || !lower_tail && p_ < p * _1_m)
                    return 0;
            } else
                p_ = (0, pgamma_1.pgamma)(x, alpha, scale, lower_tail, log_p);
            if (p_ === -Infinity)
                return 0;
            for (i = 1; i <= max_it_Newton; i++) {
                p1 = p_ - p;
                if (i === 1)
                    printer_qgamma(' it=%d: p=%d, x = %d, p.=%d; p1=d{p}=%d', i, p, x, p_, p1);
                if (i >= 2)
                    printer_qgamma('          x{it= %d} = %d, p.=%d, p1=d{p}=%d', i, x, p_, p1);
                if (Math.abs(p1) < Math.abs(EPS_N * p))
                    break;
                g = (0, dgamma_1.dgamma)(x, alpha, scale, log_p);
                if (g === (0, r_func_1.R_D__0)(log_p)) {
                    if (i === 1)
                        printer_qgamma('no final Newton step because dgamma(*)== 0!');
                    break;
                }
                t = log_p ? p1 * Math.exp(p_ - g) : p1 / g;
                t = lower_tail ? x - t : x + t;
                p_ = (0, pgamma_1.pgamma)(t, alpha, scale, lower_tail, log_p);
                if (Math.abs(p_ - p) > Math.abs(p1) || i > 1 && Math.abs(p_ - p) === Math.abs(p1)) {
                    if (i === 1 && max_it_Newton > 1)
                        printer_qgamma('no Newton step done since delta{p} >= last delta');
                    break;
                }
                if (t > 1.1 * x)
                    t = 1.1 * x;
                else if (t < 0.9 * x)
                    t = 0.9 * x;
                x = t;
            }
        }
        return x;
    };
    if (alpha < 1e-10) {
        printer_qgamma('value of shape (%d) is extremely small: results may be unreliable', alpha);
        max_it_Newton = 7;
    }
    p_ = (0, expm1_1.R_DT_qIv)(lower_tail, log_p, p);
    printer_qgamma('qgamma(p=%d, alpha=%d, scale=%d, l.t.=%s, log_p=%s): ', p, alpha, scale, lower_tail, log_p);
    g = (0, lgammafn_sign_1.lgammafn_sign)(alpha);
    ch = qchisq_appr(p, 2 * alpha, g, lower_tail, log_p, EPS1);
    if (!isFinite(ch)) {
        max_it_Newton = 0;
        return fn_end();
    }
    if (ch < EPS2) {
        max_it_Newton = 20;
        return fn_end();
    }
    if (p_ > pMAX || p_ < pMIN) {
        max_it_Newton = 20;
        return fn_end();
    }
    printer_qgamma('\t==> ch = %d:', ch);
    const c = alpha - 1;
    const s6 = (120 + c * (346 + 127 * c)) * i5040;
    const ch0 = ch;
    for (i = 1; i <= MAXIT; i++) {
        q = ch;
        p1 = 0.5 * ch;
        p2 = p_ - (0, pgamma_1.pgamma_raw)(p1, alpha, true, false);
        if (i === 1)
            printer_qgamma(' Ph.II iter; ch=%d, p2=%d', ch, p2);
        if (i >= 2)
            printer_qgamma('     it=%d,  ch=%d, p2=%d', i, ch, p2);
        if (!isFinite(p2) || ch <= 0) {
            ch = ch0;
            max_it_Newton = 27;
            return fn_end();
        }
        t = p2 * Math.exp(alpha * r_func_1.M_LN2 + g + p1 - c * Math.log(ch));
        b = t / ch;
        a = 0.5 * t - b * c;
        s1 = (210 + a * (140 + a * (105 + a * (84 + a * (70 + 60 * a))))) * i420;
        s2 = (420 + a * (735 + a * (966 + a * (1141 + 1278 * a)))) * i2520;
        s3 = (210 + a * (462 + a * (707 + 932 * a))) * i2520;
        s4 = (252 + a * (672 + 1182 * a) + c * (294 + a * (889 + 1740 * a))) * i5040;
        s5 = (84 + 2264 * a + c * (1175 + 606 * a)) * i2520;
        ch += t * (1 + 0.5 * t * s1 - b * c * (s1 - b * (s2 - b * (s3 - b * (s4 - b * (s5 - b * s6))))));
        if (Math.abs(q - ch) < EPS2 * ch) {
            return fn_end();
        }
        if (Math.abs(q - ch) > 0.1 * ch) {
            if (ch < q)
                ch = 0.9 * q;
            else
                ch = 1.1 * q;
        }
    }
    printer_qgamma('qgamma(%d) not converged in %d iterations; rel.ch=%d', p, MAXIT, ch / Math.abs(q - ch));
    return fn_end();
}
exports.qgamma = qgamma;