import { debug } from '@mangos/debug';
import {
    ML_ERR_return_NAN2,
    lineInfo4,
    ML_ERROR2,
    ME
} from '../../../packages/common/logger.mjs';
import {
    M_LN_SQRT_2PI,
    R_D__1,
    R_D_exp,
    R_DT_0,
    R_DT_1,
    R_DT_val
} from '../../r-func.mjs';
import { R_Log1_Exp } from '../exp/expm1.mjs';
import { lgammafn_sign as lgammafn } from '../../special/gamma/lgammafn_sign.mjs';
import { logspace_add } from '../gamma/logspace-add.mjs';
import { pchisq } from './pchisq.mjs';
export const DBL_MAX_EXP = Math.log2(Number.MAX_VALUE);
export const DBL_MIN_EXP = Math.log2(Number.MIN_VALUE);
const _dbl_min_exp = Math.LN2 * DBL_MIN_EXP;
const printer = debug('pnchisq');
export function pnchisq(x, df, ncp, lower_tail, log_p) {
    let ans;
    if (isNaN(x) || isNaN(df) || isNaN(ncp)) {
        return NaN;
    }
    if (!isFinite(df) || !isFinite(ncp)) {
        return ML_ERR_return_NAN2(printer, lineInfo4);
    }
    if (df < 0 || ncp < 0) {
        return ML_ERR_return_NAN2(printer, lineInfo4);
    }
    ans = pnchisq_raw(x, df, ncp, 1e-12, 8 * Number.EPSILON, 1000000, lower_tail, log_p);
    if (ncp >= 80) {
        if (lower_tail) {
            ans = Math.min(ans, R_D__1(log_p));
        } else {
            if (ans < (log_p ? -10 * Math.LN10 : 1e-10))
                ML_ERROR2(ME.ME_PRECISION, 'pnchisq', printer);
            if (!log_p)
                ans = Math.max(ans, 0);
        }
    }
    if (!log_p || ans < -1e-8) {
        return ans;
    }
    printer('   pnchisq_raw(*, log_p): ans=%d => 2nd call, other tail', ans);
    ans = pnchisq_raw(x, df, ncp, 1e-12, 8 * Number.EPSILON, 1000000, !lower_tail, false);
    return Math.log1p(-ans);
}
export function pnchisq_raw(x, f, theta, errmax, reltol, itrmax, lower_tail, log_p) {
    let term;
    let bound;
    let f_x_2n;
    let f_2n;
    let l_lam = -1;
    let l_x = -1;
    let lamSml;
    let tSml;
    let is_r;
    let is_b;
    let is_it;
    let ans;
    let u;
    let v;
    let t;
    let lt;
    let lu = -1;
    if (x <= 0) {
        if (x === 0 && f === 0) {
            const _L = -0.5 * theta;
            const result = lower_tail ? R_D_exp(log_p, _L) : log_p ? R_Log1_Exp(_L) : -Math.expm1(_L);
            printer('result1:%d', result);
            return result;
        }
        return R_DT_0(lower_tail, log_p);
    }
    if (!isFinite(x))
        return R_DT_1(lower_tail, log_p);
    if (theta < 80) {
        let ans;
        let i;
        const p1 = lgammafn(f / 2 + 1);
        const p2 = Math.LN2 + 2 / f * (p1 + _dbl_min_exp);
        const p3 = Math.log(x);
        if (lower_tail && f > 0 && p3 < p2) {
            const lambda = 0.5 * theta;
            let sum = -Infinity;
            let sum2 = -Infinity;
            let pr = -lambda;
            for (i = 0; i < 110; pr += Math.log(lambda) - Math.log(++i)) {
                sum2 = logspace_add(sum2, pr);
                sum = logspace_add(sum, pr + pchisq(x, f + 2 * i, lower_tail, true));
                if (sum2 >= -1e-15) {
                    break;
                }
            }
            ans = sum - sum2;
            printer('pnchisq(x=%d, f=%d, th.=%d); th. < 80, logspace: i=%d, ans=(sum=%d)-(sum2=%d)', x, f, theta, i, sum, sum2);
            return log_p ? ans : Math.exp(ans);
        } else {
            const lambda = 0.5 * theta;
            let sum = 0;
            let sum2 = 0;
            let pr = Math.exp(-lambda);
            for (i = 0; i < 110; pr *= lambda / ++i) {
                sum2 += pr;
                sum += pr * pchisq(x, f + 2 * i, lower_tail, false);
                if (sum2 >= 1 - 1e-15)
                    break;
            }
            ans = sum / sum2;
            printer('pnchisq(x=%d, f=%d, theta=%d); theta < 80: i=%d, sum=%d, sum2=%d', x, f, theta, i, sum, sum2);
            return log_p ? Math.log(ans) : ans;
        }
    }
    printer('pnchisq(x=%d, f=%d, theta=%d >= 80): ', x, f, theta);
    const lam = 0.5 * theta;
    lamSml = -lam < _dbl_min_exp;
    if (lamSml) {
        u = 0;
        lu = -lam;
        l_lam = Math.log(lam);
    } else {
        u = Math.exp(-lam);
    }
    v = u;
    const x2 = 0.5 * x;
    const f2 = 0.5 * f;
    f_x_2n = f - x;
    printer('-- v=exp(-th/2)=%d, x/2= %d, f/2= %d', v, x2, f2);
    if (f2 * Number.EPSILON > 0.125 && Math.abs(t = x2 - f2) < Math.sqrt(Number.EPSILON) * f2) {
        lt = (1 - t) * (2 - t / (f2 + 1)) - M_LN_SQRT_2PI - 0.5 * Math.log(f2 + 1);
        printer(' (case I) === > ');
    } else {
        lt = f2 * Math.log(x2) - x2 - lgammafn(f2 + 1);
    }
    printer(' lt= %d', lt);
    tSml = lt < _dbl_min_exp;
    if (tSml) {
        printer(' is very small');
        if (x > f + theta + 5 * Math.sqrt(2 * (f + 2 * theta))) {
            return R_DT_1(lower_tail, log_p);
        }
        l_x = Math.log(x);
        ans = term = 0;
        t = 0;
    } else {
        t = Math.exp(lt);
        printer(', t=Math.exp(lt)= %d', t);
        term = v * t;
        ans = term;
    }
    let n;
    for (n = 1, f_2n = f + 2, f_x_2n += 2;; n++, f_2n += 2, f_x_2n += 2) {
        printer(' _OL_: n=%d', n);
        if (f_x_2n > 0) {
            bound = t * x / f_x_2n;
            printer(' L10: n=%d; term= %d; bound= %d', n, term, bound);
            is_r = is_it = false;
            if ((is_b = bound <= errmax) && (is_r = term <= reltol * ans) || (is_it = n > itrmax)) {
                printer('BREAK n=%d %s; bound= %d %s, rel.err= %d %s', n, is_it ? '> itrmax' : '', bound, is_b ? '<= errmax' : '', term / ans, is_r ? '<= reltol' : '');
                break;
            }
        }
        if (lamSml) {
            lu += l_lam - Math.log(n);
            if (lu >= _dbl_min_exp) {
                printer(' n=%d; nomore underflow in u = Math.exp(lu) === > change', n);
                v = u = Math.exp(lu);
                lamSml = false;
            }
        } else {
            u *= lam / n;
            v += u;
        }
        if (tSml) {
            lt += l_x - Math.log(f_2n);
            if (lt >= _dbl_min_exp) {
                printer('  n=%d; nomore underflow in t = Math.exp(lt) === > change', n);
                t = Math.exp(lt);
                tSml = false;
            }
        } else {
            t *= x / f_2n;
        }
        if (!lamSml && !tSml) {
            term = v * t;
            ans += term;
        }
    }
    if (is_it) {
        printer('pnchisq(x=%d, ..): not converged in %d iter.', x, itrmax);
    }
    printer(' ===  L_End: n=%d; term= %d; bound=%d', n, term, bound);
    return R_DT_val(lower_tail, log_p, ans);
}