import { debug } from '@mangos/debug';
import {
    ML_ERR_return_NAN2,
    lineInfo4
} from '../../../packages/common/logger.mjs';
import {
    R_D__0,
    R_D__1,
    R_D_exp,
    R_DT_0,
    R_DT_1,
    R_P_bounds_01,
    DBL_MAX_EXP,
    M_LN2,
    DBL_MIN
} from '../../r-func.mjs';
import { R_Log1_Exp } from '../exp/expm1.mjs';
import { dnorm4 as dnorm } from '../normal/dnorm.mjs';
import { pnorm5 as pnorm } from '../normal/pnorm.mjs';
import { dpois_raw } from '../poisson/dpois.mjs';
import { lgammafn_sign as lgammafn } from '../../special/gamma/lgammafn_sign.mjs';
const scalefactor = 4294967296 ** 8;
const M_cutoff = M_LN2 * DBL_MAX_EXP / Number.EPSILON;
function logcf(x, i, d, eps) {
    let c1 = 2 * d;
    let c2 = i + d;
    let c4 = c2 + d;
    let a1 = c2;
    let b1 = i * (c2 - i * x);
    let b2 = d * d * x;
    let a2 = c4 * c2 - b2;
    b2 = c4 * b1 - i * b2;
    while (Math.abs(a2 * b1 - a1 * b2) > Math.abs(eps * b1 * b2)) {
        let c3 = c2 * c2 * x;
        c2 += d;
        c4 += d;
        a1 = c4 * a2 - c3 * a1;
        b1 = c4 * b2 - c3 * b1;
        c3 = c1 * c1 * x;
        c1 += d;
        c4 += d;
        a2 = c4 * a1 - c3 * a2;
        b2 = c4 * b1 - c3 * b2;
        if (Math.abs(b2) > scalefactor) {
            a1 /= scalefactor;
            b1 /= scalefactor;
            a2 /= scalefactor;
            b2 /= scalefactor;
        } else if (Math.abs(b2) < 1 / scalefactor) {
            a1 *= scalefactor;
            b1 *= scalefactor;
            a2 *= scalefactor;
            b2 *= scalefactor;
        }
    }
    return a2 / b2;
}
function log1pmx(x) {
    const minLog1Value = -0.79149064;
    if (x > 1 || x < minLog1Value)
        return Math.log1p(x) - x;
    else {
        const r = x / (2 + x);
        const y = r * r;
        if (Math.abs(x) < 0.01) {
            const two = 2;
            return r * ((((two / 9 * y + two / 7) * y + two / 5) * y + two / 3) * y - x);
        } else {
            const tol_logcf = 1e-14;
            return r * (2 * y * logcf(y, 3, 2, tol_logcf) - x);
        }
    }
}
const coeffs = new Float64Array([
    0.3224670334241132,
    0.0673523010531981,
    0.020580808427784546,
    0.007385551028673986,
    0.0028905103307415234,
    0.001192753911703261,
    0.0005096695247430425,
    0.00022315475845357939,
    0.00009945751278180853,
    0.00004492623673813314,
    0.00002050721277567069,
    0.000009439488275268397,
    0.000004374866789907488,
    0.000002039215753801366,
    9.55141213040742e-7,
    4.492469198764566e-7,
    2.1207184805554665e-7,
    1.0043224823968099e-7,
    4.7698101693639804e-8,
    2.2711094608943164e-8,
    1.0838659214896955e-8,
    5.183475041970047e-9,
    2.4836745438024785e-9,
    1.1921401405860912e-9,
    5.731367241678862e-10,
    2.7595228851242334e-10,
    1.330476437424449e-10,
    6.4229645638381e-11,
    3.1044247747322276e-11,
    1.5021384080754142e-11,
    7.275974480239079e-12,
    3.527742476575915e-12,
    1.711991790559618e-12,
    8.315385841420285e-13,
    4.04220052528944e-13,
    1.9664756310966165e-13,
    9.573630387838556e-14,
    4.6640760264283744e-14,
    2.2737369600659724e-14,
    1.1091399470834522e-14
]);
export function lgamma1p(a) {
    const eulers_const = 0.5772156649015329;
    const N = 40;
    const c = 2.2737368458246524e-13;
    const tol_logcf = 1e-14;
    let lgam;
    let i;
    if (Math.abs(a) >= 0.5)
        return lgammafn(a + 1);
    lgam = c * logcf(-a / 2, N + 2, 1, tol_logcf);
    for (i = N - 1; i >= 0; i--)
        lgam = coeffs[i] - a * lgam;
    return (a * lgam - eulers_const) * a - log1pmx(a);
}
const pr_dpois_wrap = debug('dpois_wrap');
function dpois_wrap(x_plus_1, lambda, give_log) {
    pr_dpois_wrap('dpois_wrap(x+1=%d, lambda=%d, log=%s)', x_plus_1, lambda, give_log);
    if (!isFinite(lambda)) {
        return R_D__0(give_log);
    }
    if (x_plus_1 > 1)
        return dpois_raw(x_plus_1 - 1, lambda, give_log);
    if (lambda > Math.abs(x_plus_1 - 1) * M_cutoff)
        return R_D_exp(give_log, -lambda - lgammafn(x_plus_1));
    else {
        const d = dpois_raw(x_plus_1, lambda, give_log);
        pr_dpois_wrap('  -> d=dpois_raw(..)=%d', d);
        return give_log ? d + Math.log(x_plus_1 / lambda) : d * (x_plus_1 / lambda);
    }
}
const pr_pgamma_smallx = debug('pgamma_smallx');
function pgamma_smallx(x, alph, lowerTail, logP) {
    let sum = 0;
    let c = alph;
    let n = 0;
    let term;
    pr_pgamma_smallx(' pg_smallx(x=%d, alph=%d): ', x, alph);
    do {
        n++;
        c *= -x / n;
        term = c / (alph + n);
        sum += term;
    } while (Math.abs(term) > Number.EPSILON * Math.abs(sum));
    pr_pgamma_smallx('%d terms --> conv.sum=%d;', n, sum);
    if (lowerTail) {
        const f1 = logP ? Math.log1p(sum) : 1 + sum;
        let f2;
        if (alph > 1) {
            f2 = dpois_raw(alph, x, logP);
            f2 = logP ? f2 + x : f2 * Math.exp(x);
        } else if (logP) {
            f2 = alph * Math.log(x) - lgamma1p(alph);
        } else {
            f2 = Math.pow(x, alph) / Math.exp(lgamma1p(alph));
        }
        pr_pgamma_smallx(' (f1,f2)= (%d,%d)', f1, f2);
        return logP ? f1 + f2 : f1 * f2;
    } else {
        const lf2 = alph * Math.log(x) - lgamma1p(alph);
        pr_pgamma_smallx(' 1:%d  2:%d', alph * Math.log(x), lgamma1p(alph));
        pr_pgamma_smallx(' sum=%d  log(1+sum)=%d\t lf2=%d', sum, Math.log1p(sum), lf2);
        if (logP)
            return R_Log1_Exp(Math.log1p(sum) + lf2);
        else {
            const f1m1 = sum;
            const f2m1 = Math.expm1(lf2);
            return -(f1m1 + f2m1 + f1m1 * f2m1);
        }
    }
}
function pd_upper_series(x, y, logP) {
    let term = x / y;
    let sum = term;
    do {
        y++;
        term *= x / y;
        sum += term;
    } while (term > sum * Number.EPSILON);
    return logP ? Math.log(sum) : sum;
}
const pr_pd_lower_cf = debug('pd_lower_cf');
function pd_lower_cf(y, d) {
    let f = 0;
    let of;
    let f0;
    let i;
    let c2;
    let c3;
    let c4;
    let a1;
    let b1;
    let a2;
    let b2;
    const max_it = 200000;
    pr_pd_lower_cf('pd_lower_cf(y=%d, d=%d)', y, d);
    if (y === 0)
        return 0;
    f0 = y / d;
    if (Math.abs(y - 1) < Math.abs(d) * Number.EPSILON) {
        pr_pd_lower_cf(' very small "y" -> returning (y/d)');
        return f0;
    }
    if (f0 > 1)
        f0 = 1;
    c2 = y;
    c4 = d;
    a1 = 0;
    b1 = 1;
    a2 = y;
    b2 = d;
    while (b2 > scalefactor) {
        a1 /= scalefactor;
        b1 /= scalefactor;
        a2 /= scalefactor;
        b2 /= scalefactor;
    }
    i = 0;
    of = -1;
    while (i < max_it) {
        i++;
        c2--;
        c3 = i * c2;
        c4 += 2;
        a1 = c4 * a2 + c3 * a1;
        b1 = c4 * b2 + c3 * b1;
        i++;
        c2--;
        c3 = i * c2;
        c4 += 2;
        a2 = c4 * a1 + c3 * a2;
        b2 = c4 * b1 + c3 * b2;
        if (b2 !== 0) {
            f = a2 / b2;
            if (Math.abs(f - of) <= Number.EPSILON * Math.max(f0, Math.abs(f))) {
                pr_pd_lower_cf(' %d iter.\n', i);
                return f;
            }
            of = f;
        }
    }
    pr_pd_lower_cf(' ** NON-convergence in pgamma()\'s pd_lower_cf() f= %d.', f);
    return f;
}
const pr_pd_lower_series = debug('pd_lower_series');
function pd_lower_series(lambda, y) {
    let term = 1;
    let sum = 0;
    pr_pd_lower_series('pd_lower_series(lam=%d, y=%d) ...', lambda, y);
    while (y >= 1 && term > sum * Number.EPSILON) {
        term *= y / lambda;
        sum += term;
        y--;
    }
    pr_pd_lower_series(' done: term=%d, sum=%d, y= %d', term, sum, y);
    if (y !== Math.floor(y)) {
        pr_pd_lower_series(' y not int: add another term ');
        const f = pd_lower_cf(y, lambda + 1 - y);
        pr_pd_lower_series('  (= %d) * term = %d to sum %d', f, term * f, sum);
        sum += term * f;
    }
    return sum;
}
function dpnorm(x, lowerTail, lp) {
    if (x < 0) {
        x = -x;
        lowerTail = !lowerTail;
    }
    if (x > 10 && !lowerTail) {
        let term = 1 / x;
        let sum = term;
        const x2 = x * x;
        let i = 1;
        do {
            term *= -i / x2;
            sum += term;
            i += 2;
        } while (Math.abs(term) > Number.EPSILON * sum);
        return 1 / sum;
    } else {
        const d = dnorm(x, 0, 1, false);
        return d / Math.exp(lp);
    }
}
const pr_ppois_asymp = debug('ppois_asymp');
const coefs_a = new Float64Array([
    -1e+99,
    2 / 3,
    -4 / 135,
    8 / 2835,
    16 / 8505,
    -8992 / 12629925,
    -334144 / 492567075,
    698752 / 1477701225
]);
const coefs_b = new Float64Array([
    -1e+99,
    1 / 12,
    1 / 288,
    -139 / 51840,
    -571 / 2488320,
    163879 / 209018880,
    5246819 / 75246796800,
    -534703531 / 902961561600
]);
function ppois_asymp(x, lambda, lowerTail, logP) {
    let elfb;
    let elfb_term;
    let res12;
    let res1_term;
    let res1_ig;
    let res2_term;
    let res2_ig;
    let s2pt;
    let i;
    const dfm = lambda - x;
    const pt_ = -log1pmx(dfm / x);
    s2pt = Math.sqrt(2 * x * pt_);
    if (dfm < 0)
        s2pt = -s2pt;
    res12 = 0;
    res1_ig = res1_term = Math.sqrt(x);
    res2_ig = res2_term = s2pt;
    for (i = 1; i < 8; i++) {
        res12 += res1_ig * coefs_a[i];
        res12 += res2_ig * coefs_b[i];
        res1_term *= pt_ / i;
        res2_term *= 2 * pt_ / (2 * i + 1);
        res1_ig = res1_ig / x + res1_term;
        res2_ig = res2_ig / x + res2_term;
    }
    elfb = x;
    elfb_term = 1;
    for (i = 1; i < 8; i++) {
        elfb += elfb_term * coefs_b[i];
        elfb_term /= x;
    }
    if (!lowerTail)
        elfb = -elfb;
    pr_ppois_asymp('res12 = %d   elfb=%d', elfb, res12);
    const f = res12 / elfb;
    const np = pnorm(s2pt, 0, 1, !lowerTail, logP);
    if (logP) {
        const n_d_over_p = dpnorm(s2pt, !lowerTail, np);
        pr_ppois_asymp('pp*_asymp(): f=%d\t np=e^%d  nd/np=%d  f*nd/np=%d', f, np, n_d_over_p, f * n_d_over_p);
        return np + Math.log1p(f * n_d_over_p);
    } else {
        const nd = dnorm(s2pt, 0, 1, logP);
        pr_ppois_asymp('pp*_asymp(): f=%d\t np=%d  nd=%d  f*nd=%d', f, np, nd, f * nd);
        return np + f * nd;
    }
}
const pr_pgamma_raw = debug('pgamma_raw');
export function pgamma_raw(x, alph, lowerTail, logP) {
    let res;
    pr_pgamma_raw('pgamma_raw(x=%d, alph=%d, low=%s, log=%s)', x, alph, lowerTail, logP);
    const rc = R_P_bounds_01(lowerTail, logP, x, 0, Infinity);
    if (rc !== undefined) {
        return rc;
    }
    if (x < 1) {
        res = pgamma_smallx(x, alph, lowerTail, logP);
    } else if (x <= alph - 1 && x < 0.8 * (alph + 50)) {
        const sum = pd_upper_series(x, alph, logP);
        const d = dpois_wrap(alph, x, logP);
        pr_pgamma_raw(' alph "large": sum=pd_upper*()= %d, d=dpois_w(*)= %d', sum, d);
        if (!lowerTail) {
            res = logP ? R_Log1_Exp(d + sum) : 1 - d * sum;
        } else {
            res = logP ? sum + d : sum * d;
        }
    } else if (alph - 1 < x && alph < 0.8 * (x + 50)) {
        let sum;
        const d = dpois_wrap(alph, x, logP);
        pr_pgamma_raw('  x "large": d=dpois_w(*)= %d ', d);
        if (alph < 1) {
            if (x * Number.EPSILON > 1 - alph)
                sum = R_D__1(logP);
            else {
                const f = pd_lower_cf(alph, x - (alph - 1)) * x / alph;
                sum = logP ? Math.log(f) : f;
            }
        } else {
            sum = pd_lower_series(x, alph - 1);
            sum = logP ? Math.log1p(sum) : 1 + sum;
        }
        pr_pgamma_raw(', sum= %d', sum);
        if (!lowerTail)
            res = logP ? sum + d : sum * d;
        else
            res = logP ? R_Log1_Exp(d + sum) : 1 - d * sum;
    } else {
        pr_pgamma_raw(' using ppois_asymp()');
        res = ppois_asymp(alph - 1, x, !lowerTail, logP);
    }
    if (!logP && res < DBL_MIN / Number.EPSILON) {
        pr_pgamma_raw(' very small res=%.14g; -> recompute via log\n', res);
        return Math.exp(pgamma_raw(x, alph, lowerTail, true));
    } else
        return res;
}
const printer_pgamma = debug('pgamma');
export function pgamma(x, shape, scale, lowerTail, logP) {
    if (isNaN(x) || isNaN(shape) || isNaN(scale)) {
        return NaN;
    }
    if (shape < 0 || scale <= 0)
        return ML_ERR_return_NAN2(printer_pgamma, lineInfo4);
    x /= scale;
    if (shape === 0)
        return x <= 0 ? R_DT_0(lowerTail, logP) : R_DT_1(lowerTail, logP);
    return pgamma_raw(x, shape, lowerTail, logP);
}