'use strict';
import { debug } from '@mangos/debug';
import {
    ME,
    ML_ERROR2,
    R_Q_P01_boundaries,
    ML_ERR_return_NAN2,
    lineInfo4
} from '../../../packages/common/logger.mjs';
import {
    DBL_MANT_DIG,
    M_1_PI,
    M_PI_2,
    R_D_Cval,
    R_D_log,
    R_D_Lval,
    R_D_qIv,
    DBL_EPSILON,
    min,
    abs,
    expm1,
    sqrt,
    exp,
    LN2,
    SQRT2,
    pow,
    PI,
    log,
    DBL_MAX,
    DBL_MIN
} from '../../r-func.mjs';
import {
    R_D_LExp,
    R_DT_qIv
} from '../exp/expm1.mjs';
import { qnorm } from '../normal/qnorm.mjs';
import { tanpi } from '../../trigonometry/tanpi.mjs';
import { _dt } from './dt.mjs';
import { pt } from './pt.mjs';
const printer = debug('qt');
const accu = 1e-13;
const Eps = 1e-11;
const eps = 1e-12;
export function qt(p, ndf, lower_tail, log_p) {
    let P;
    let q;
    if (isNaN(p) || isNaN(ndf)) {
        return p + ndf;
    }
    const rc = R_Q_P01_boundaries(lower_tail, log_p, p, -Infinity, Infinity);
    if (rc !== undefined) {
        return rc;
    }
    if (ndf <= 0) {
        return ML_ERR_return_NAN2(printer, lineInfo4);
    }
    if (ndf < 1) {
        let ux;
        let lx;
        let nx;
        let pp;
        let iter = 0;
        p = R_DT_qIv(lower_tail, log_p, p);
        if (p > 1 - DBL_EPSILON) {
            return Infinity;
        }
        pp = min(1 - DBL_EPSILON, p * (1 + Eps));
        for (ux = 1; ux < DBL_MAX && pt(ux, ndf, true, false) < pp; ux *= 2);
        pp = p * (1 - Eps);
        for (lx = -1; lx > -DBL_MAX && pt(lx, ndf, true, false) > pp; lx *= 2);
        do {
            nx = 0.5 * (lx + ux);
            if (pt(nx, ndf, true, false) > p) {
                ux = nx;
            } else {
                lx = nx;
            }
        } while ((ux - lx) / abs(nx) > accu && ++iter < 1000);
        if (iter >= 1000) {
            ML_ERROR2(ME.ME_PRECISION, 'qt', printer);
        }
        return 0.5 * (lx + ux);
    }
    if (ndf > 100000000000000000000) {
        return qnorm(p, 0, 1, lower_tail, log_p);
    }
    P = R_D_qIv(log_p, p);
    const neg = (!lower_tail || P < 0.5) && (lower_tail || P > 0.5);
    const is_neg_lower = lower_tail === neg;
    if (neg) {
        if (log_p) {
            if (lower_tail) {
                P = 2 * P;
            } else {
                P = 2 * -expm1(p);
            }
        } else {
            P = 2 * R_D_Lval(lower_tail, p);
        }
    } else {
        if (log_p) {
            if (lower_tail) {
                P = 2 * -expm1(p);
            } else {
                P = 2 * P;
            }
        } else {
            P = 2 * R_D_Cval(lower_tail, p);
        }
    }
    if (abs(ndf - 2) < eps) {
        if (P > DBL_MIN) {
            if (3 * P < DBL_EPSILON) {
                q = 1 / sqrt(P);
            } else if (P > 0.9) {
                q = (1 - P) * sqrt(2 / (P * (2 - P)));
            } else {
                q = sqrt(2 / (P * (2 - P)) - 2);
            }
        } else {
            if (log_p) {
                q = is_neg_lower ? exp(-p / 2) / SQRT2 : 1 / sqrt(-expm1(p));
            } else {
                q = Infinity;
            }
        }
    } else if (ndf < 1 + eps) {
        if (P === 1) {
            q = 0;
        } else if (P > DBL_EPSILON / 2) {
            q = 1 / tanpi(P / 2);
        } else {
            if (log_p) {
                q = is_neg_lower ? M_1_PI * exp(-p) : -1 / (PI * expm1(p));
            } else {
                q = Infinity;
            }
        }
    } else {
        let x = 0;
        let y = 0;
        let log_P2 = 0;
        const a = 1 / (ndf - 0.5);
        const b = 48 / (a * a);
        let c = ((20700 * a / b - 98) * a - 16) * a + 96.36;
        const d = ((94.5 / (b + c) - 3) / b + 1) * sqrt(a * M_PI_2) * ndf;
        const P_ok1 = P > DBL_MIN || !log_p;
        let P_ok = P_ok1;
        if (P_ok1) {
            y = pow(d * P, 2 / ndf);
            P_ok = y >= DBL_EPSILON;
        }
        if (!P_ok) {
            if (is_neg_lower) {
                log_P2 = R_D_log(log_p, p);
            } else {
                log_P2 = R_D_LExp(log_p, p);
            }
            x = (log(d) + LN2 + log_P2) / ndf;
            y = exp(2 * x);
        }
        if (ndf < 2.1 && P > 0.5 || y > 0.05 + a) {
            if (P_ok) {
                x = qnorm(0.5 * P, 0, 1, false, false);
            } else {
                x = qnorm(log_P2, 0, 1, lower_tail, true);
            }
            y = x * x;
            if (ndf < 5) {
                c += 0.3 * (ndf - 4.5) * (x + 0.6);
            }
            c = (((0.05 * d * x - 5) * x - 7) * x - 2) * x + b + c;
            y = (((((0.4 * y + 6.3) * y + 36) * y + 94.5) / c - y - 3) / b + 1) * x;
            y = expm1(a * y * y);
            q = sqrt(ndf * y);
        } else if (!P_ok && x < -LN2 * DBL_MANT_DIG) {
            q = sqrt(ndf) * exp(-x);
        } else {
            y = ((1 / (((ndf + 6) / (ndf * y) - 0.089 * d - 0.822) * (ndf + 2) * 3) + 0.5 / (ndf + 4)) * y - 1) * (ndf + 1) / (ndf + 2) + 1 / y;
            q = sqrt(ndf * y);
        }
        if (P_ok1) {
            let it = 0;
            while (it++ < 10 && (y = _dt(q, ndf, false)) > 0 && isFinite(x = (pt(q, ndf, false, false) - P / 2) / y) && abs(x) > 1e-14 * abs(q)) {
                q += x * (1 + x * q * (ndf + 1) / (2 * (q * q + ndf)));
            }
        }
    }
    if (neg) {
        q = -q;
    }
    return q;
}