import { KnuthTAOCP } from './knuth-taocp/index.mjs';
import { KnuthTAOCP2002 } from './knuth-taocp-2002/index.mjs';
import { LecuyerCMRG } from './lecuyer-cmrg/index.mjs';
import { MarsagliaMultiCarry } from './marsaglia-multicarry/index.mjs';
import { MersenneTwister } from './mersenne-twister/index.mjs';
import { SuperDuper } from './super-duper/index.mjs';
import { WichmannHill } from './wichmann-hill/index.mjs';
import { AhrensDieter } from './normal/ahrens-dieter/index.mjs';
import { BoxMuller } from './normal/box-muller/index.mjs';
import { BuggyKindermanRamage } from './normal/buggy-kinderman-ramage/index.mjs';
import { Inversion } from './normal/inversion/index.mjs';
import { KindermanRamage } from './normal/kinderman-ramage/index.mjs';
const uniformMap = {
    KNUTH_TAOCP: KnuthTAOCP,
    KNUTH_TAOCP2002: KnuthTAOCP2002,
    LECUYER_CMRG: LecuyerCMRG,
    MARSAGLIA_MULTICARRY: MarsagliaMultiCarry,
    MERSENNE_TWISTER: MersenneTwister,
    SUPER_DUPER: SuperDuper,
    WICHMANN_HILL: WichmannHill
};
const normalMap = {
    AHRENS_DIETER: AhrensDieter,
    BOX_MULLER: BoxMuller,
    BUGGY_KINDERMAN_RAMAGE: BuggyKindermanRamage,
    INVERSION: Inversion,
    KINDERMAN_RAMAGE: KindermanRamage
};
const uniformKeys = {
    KNUTH_TAOCP: 'KNUTH_TAOCP',
    KNUTH_TAOCP2002: 'KNUTH_TAOCP2002',
    LECUYER_CMRG: 'LECUYER_CMRG',
    MARSAGLIA_MULTICARRY: 'MARSAGLIA_MULTICARRY',
    MERSENNE_TWISTER: 'MERSENNE_TWISTER',
    SUPER_DUPER: 'SUPER_DUPER',
    WICHMANN_HILL: 'WICHMANN_HILL'
};
const normalKeys = {
    AHRENS_DIETER: 'AHRENS_DIETER',
    BOX_MULLER: 'BOX_MULLER',
    BUGGY_KINDERMAN_RAMAGE: 'BUGGY_KINDERMAN_RAMAGE',
    INVERSION: 'INVERSION',
    KINDERMAN_RAMAGE: 'KINDERMAN_RAMAGE'
};
const sampleKind = {
    ROUNDING: 'ROUNDING',
    REJECTION: 'REJECTION'
};
const symRNG = Symbol.for('rngUNIFORM');
const symRNGNormal = Symbol.for('rngNORMAL');
const symSampleKind = Symbol.for('sample.kind');
export function globalUni(d) {
    if (d) {
        globalThis[symRNG] = d;
    }
    return globalThis[symRNG];
}
export function globalNorm(d) {
    if (d) {
        globalThis[symRNGNormal] = d;
    }
    return globalThis[symRNGNormal];
}
export function globalSampleKind(d) {
    if (d) {
        globalThis[symSampleKind] = d;
    }
    return globalThis[symSampleKind];
}
export function setSeed(seed, randomSet) {
    if (typeof seed !== 'number') {
        throw new Error('Seed needs to bea number');
    }
    if (isNaN(seed) || !isFinite(seed)) {
        throw new Error('Seed needs to be a finite number');
    }
    if (randomSet) {
        RNGkind(randomSet);
    }
    const gu = globalUni();
    const s = new Uint32Array([0]);
    s[0] = seed;
    if (s[0] !== seed) {
        throw new Error(`Seed needs to be an unsigned 32 bit integer, it is ${ seed }`);
    }
    gu.init(seed);
}
function RNGkind(opt = {}) {
    let gu = globalUni();
    let no = globalNorm();
    let sk = globalSampleKind();
    function testAndSetUniform(u) {
        const tu = uniformMap[u];
        if (tu) {
            if (tu.kind !== gu.constructor.kind) {
                no.unregister();
                gu = globalUni(new tu());
                no.register(gu);
            }
            return true;
        }
        return false;
    }
    function testAndSetNormal(n) {
        const tn = normalMap[n];
        if (tn) {
            if (tn.kind !== no.constructor.kind) {
                no.unregister();
                no = globalNorm(new tn(gu));
                no.register(gu);
            }
            return true;
        }
        return false;
    }
    function testAndSetSampleKind(s) {
        if (s === 'REJECTION' || s === 'ROUNDING') {
            if (s !== sk) {
                sk = globalSampleKind(s);
            }
            return true;
        }
        return false;
    }
    if (opt.uniform) {
        testAndSetUniform(opt.uniform);
    }
    if (opt.normal) {
        testAndSetNormal(opt.normal);
    }
    if (opt.sampleKind) {
        testAndSetSampleKind(opt.sampleKind);
    }
    return {
        uniform: gu.constructor.kind,
        normal: no.constructor.kind,
        sampleKind: sk
    };
}
RNGkind.uniform = uniformKeys;
RNGkind.normal = normalKeys;
RNGkind.sampleKind = sampleKind;
export {
    RNGkind
};
export function randomSeed(internalState) {
    const gu = globalUni();
    if (internalState === undefined) {
        return gu.seed;
    }
    const state = gu.seed;
    if (state.constructor.name === internalState.constructor.name && state.length === internalState.length) {
        gu.seed = internalState;
    }
    throw new Error(`the internal state of ${ gu.name } is a ${ state.constructor.name } of length ${ state.length }`);
}
globalThis[symRNG] = new MersenneTwister();
globalThis[symRNGNormal] = new Inversion(globalThis[symRNG]);
globalThis[symSampleKind] = 'ROUNDING';